;;; ns-mark-nav.el -- mark navigation commands for Emacs

;; Copyright (C) 2005
;;        Free Software Foundation, Inc.

;; Author: Adrian Robert (arobert@cogsci.ucsd.edu)

;; This file is part of Emacs.app

;;; Commentary:

;; This is some experimental lisp to support moving backwards and forwards
;; in the so-called "mark ring".  Since the mark ring is in reality a stack,
;; this entails augmenting that with an opposing stack to hold popped values.
;; The lisp here implements that.  (Some code lifted from simple.el.)

;; After you load this file, the keys M-p and M-n will move forwards and
;; backwards through the mark history respectively.

; Variables
(defvar alt-mark-ring nil
  "Alternate stack of marks of the current buffer, allowing mark ring to actually act like a ring and not a stack.")
(make-variable-buffer-local 'alt-mark-ring)
(put 'alt-mark-ring 'permanent-local t)


;; Helper functions
(defun pop-alt-mark ()
  "Pop off alternate mark ring into the buffer's actual mark.
Does not set point.  Does nothing if alternate mark ring is empty."
  (if alt-mark-ring
      (progn
	(setq alt-mark-ring (nconc alt-mark-ring (list (copy-marker (mark-marker)))))
	(set-marker (mark-marker) (+ 0 (car alt-mark-ring)) (current-buffer))
	(deactivate-mark)
	(move-marker (car alt-mark-ring) nil)
	(if (null (mark t)) (ding))
	(setq alt-mark-ring (cdr alt-mark-ring)))))

(defun push-alt-mark ()
  "Push current mark on alternate mark ring (alt-mark-ring) without changing."
  (if (null (mark t))
      nil
    (setq alt-mark-ring (cons (copy-marker (mark-marker)) alt-mark-ring))))

(defun push-mark-no-set ()
  "Pushes mark onto mark ring without resetting it."
  (if (null (mark t))
      nil
    (setq mark-ring (cons (copy-marker (mark-marker)) mark-ring))
    (if (> (length mark-ring) mark-ring-max)
	  (setcdr (nthcdr (1- mark-ring-max) mark-ring) nil))))


; Main functions
(defun ns-prev-mark ()
  "Move to previous mark, similar to C-u C-spc, but save the popped mark for later re-traversal -- see ns-next-mark."
  (interactive)
  (if (null (mark t))
      (error "No mark set in this buffer")
    (progn 
      (goto-char (mark t))
      (push-alt-mark)
      (pop-mark))))

(defun ns-next-mark ()
  "Move to next mark, saving current mark for re-traversal -- see ns-prev-mark."
  (interactive nil)
  (if (or (null alt-mark-ring) (null (mark t)))
      (error "No next-mark set in this buffer")
    (push-mark-no-set)
    (pop-alt-mark)
    (goto-char (mark t))))


; Keybindings
(global-set-key [(meta p)] 'ns-prev-mark)
(global-set-key [(meta n)] 'ns-next-mark)

(provide 'ns-mark-nav)

;;; mark-nav.el ends here
