/* Image support for the NeXT/Open/GNUstep and MacOSX window system.

   Copyright (C) 1989, 1992, 1993, 1994, 2005, 2006 Free Software Foundation.

This file is part of GNU Emacs.

GNU Emacs is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.

Originally by Carl Edman
Updated by Christian Limpach (chris@nice.ch)
OpenStep/Rhapsody port by Scott Bender (sbender@harmony-ds.com)
MacOSX/Aqua port by Christophe de Dinechin (descubes@earthlink.net)
GNUstep port and post-20 update by Adrian Robert (arobert@cogsci.ucsd.edu)

*/

#include "config.h"
#include "lisp.h"
#include "dispextern.h"
#include "nsterm.h"
#include "frame.h"

extern Lisp_Object QCfile, QCdata;

// call tracing
#if 0
int image_trace_num = 0;
#define NSTRACE(x)        fprintf(stderr, "%s:%d: [%d] " #x "\n",         \
                                __FILE__, __LINE__, ++image_trace_num)
#else
#define NSTRACE(x)
#endif


// ============================================================================
//
//   C interface.  This allows easy calling from C files.  We could just
//   compile everything as Objective-C, but that might mean slower
//   compilation and possible difficulties on some platforms..
//
// ============================================================================

void *
ns_image_from_XBM (char *bits, int width, int height)
{
  NSTRACE(ns_image_from_XBM);
  return [[EmacsImage alloc] initFromXBM: bits
                                   width: width height: height
                                    flip: YES];
}

void *
ns_image_for_XPM (int width, int height, int depth)
{
  NSTRACE(ns_image_for_XPM);
  id img = [[EmacsImage alloc] initForXPMWithDepth: depth
                                           width: width height: height];
}

void *
ns_image_from_file (Lisp_Object file)
{
  NSTRACE(ns_image_from_bitmap_file);
  return [EmacsImage allocInitFromFile: file];
}

int
ns_load_image (struct frame *f, struct image *img,
               Lisp_Object spec_file, Lisp_Object spec_data)
{
  NSTRACE(ns_load_image);

  EmacsImage *eImg;

  if (NILP (spec_data))
    {
      eImg = [EmacsImage allocInitFromFile: spec_file];
    }
  else
    {
      NSData *data = [NSData dataWithBytes: XSTRING (spec_data)->data
                                    length: SBYTES(spec_data)];
      eImg = [[EmacsImage alloc] initWithData: data];
    }

  if (eImg == nil)
    {
      add_to_log("Unable to load image %s", img->spec, Qnil);
      return 0;
    }

  img->width = ns_image_width(eImg);
  img->height = ns_image_height(eImg);

  // 4) set img->pixmap = emacsimage
  img->pixmap = eImg;
  return 1;
}


int
ns_image_width (void *img)
{
    return [[(id)img stippleRep] size].width;
}

int
ns_image_height (void *img)
{
    return [[(id)img stippleRep] size].height;
}

unsigned long
ns_get_pixel(void *img, int x, int y)
{
  return [(EmacsImage *)img getPixelAtX: x Y: y];
}

void
ns_put_pixel(void *img, int x, int y, void *color)
{
  [(EmacsImage *)img setPixelAtX: x Y: y to: color];
}

void
ns_set_alpha(void *img, int x, int y, int a)
{
  [(EmacsImage *)img setAlphaAtX: x Y: y to: a];
}

int ns_get_red(void *color)
{
  return [(NSColor *)color redComponent] * 65535;
}

int ns_get_green(void *color)
{
  return [(NSColor *)color greenComponent] * 65535;
}

int ns_get_blue(void *color)
{
  return [(NSColor *)color blueComponent] * 65535;
}


// ============================================================================
//
//   Class supporting bitmaps and images of various sorts.
//
// ============================================================================

@implementation EmacsImage

static EmacsImage *ImageList = nil;

+ allocInitFromFile:(Lisp_Object)file
{
  EmacsImage *image = ImageList;
  Lisp_Object found;
  int fd;

  // look for an existing image of the same name
  while (image != nil &&
         [[image name] compare: [NSString stringWithUTF8String:
                                              XSTRING (file)->data]])
    image = [image imageListNext];

  if (image != nil)
    {
      [image reference];
      return image;
    }

  /* Search bitmap-file-path for the file, if appropriate.  */
  found = x_find_image_file(file);
  if (!STRINGP(file))
    return nil;

  image = [[EmacsImage alloc] initByReferencingFile:
                     [NSString stringWithUTF8String:XSTRING (found)->data]];

  if ([image bestRepresentationForDevice:nil] == nil)
    {
      [image release];
      return nil;
    }

  [image setName:[NSString stringWithUTF8String:XSTRING (file)->data]];
  [image reference];
  ImageList = [image imageListSetNext:ImageList];

  return image;
}


- reference
{
  refCount++;
  return self;
}


- imageListSetNext:(id)arg
{
  imageListNext = arg;
  return self;
}


- imageListNext
{
  return imageListNext;
}


- (void)dealloc
{
  id list = ImageList;

  if (refCount > 1)
    {
      refCount--;
      return;
    }

  if (list == self)
    ImageList = imageListNext;
  else
    {
      while (list != nil && [list imageListNext] != self)
        list = [list imageListNext];
      [list imageListSetNext:imageListNext];
    }
  if (stippleRep && stippleRep != self)
    [stippleRep release];
  { [super dealloc]; return; };
}


- initFromXBM:(unsigned char *)bits width:(int)w height:(int)h
         flip:(BOOL)flip
{
  return [self initFromSkipXBM:bits width:w height:h flip:flip length:0];
}


- initFromSkipXBM:(unsigned char *)bits width:(int)w height:(int)h
             flip:(BOOL)flip length:(int)length;
{
  NSSize s = {w, h};
  id rep;
  int bpr = (w + 7) / 8;
  unsigned char *planes[5];

  [self initWithSize:s];

  rep = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes: NULL
                                  pixelsWide: w pixelsHigh: h
                                  bitsPerSample: 1 samplesPerPixel: 2
                                  hasAlpha:YES isPlanar:YES
                                  colorSpaceName: NSCalibratedBlackColorSpace
                                  bytesPerRow: bpr bitsPerPixel: 0];

  [rep getBitmapDataPlanes: planes];
  {
    // turn the bytes around and invert for alpha on plane 2
    int i, j;
    unsigned char *s = bits;
    unsigned char *d = planes[0];
    unsigned char *d2 = planes[1];
    unsigned char swt[16] = {0, 8, 4, 12, 2, 10, 6, 14, 1, 9, 5, 13,
                             3, 11, 7, 15};
    unsigned char c;

    for (j = 0; j < h; j++)
      for (i = 0; i < bpr; i++)
        {
          if (length)
            {
              unsigned char s1, s2;
              while (*s++ != 'x' && s < bits + length);
              if (s >= bits + length)
                {
                  [rep release];
                  return nil;
                }
#define hexchar(x) (isdigit (x) ? x - '0' : x - 'a' + 10)
              s1 = *s++;
              s2 = *s++;
              c = hexchar (s1) * 0x10 + hexchar (s2);
            }
          else
            c = *s++;

          *d = flip ? swt[c >> 4] | (swt[c & 0xf] << 4)
                    : c ^ 255;
          *d2++ = *d++;
        }
  }

  [self addRepresentation:rep];
  return self;
}


// since 48 is a popular pixmap size, 96 is a good choice
#define STIPPLE_MAX_WIDTH 96
#define STIPPLE_MAX_HEIGHT 96
- prepareForStippling
{
  NSSize r;

  if (stippleRep && stippleRep != self)
    [stippleRep release];

  r = [self size];
  if ((r.width < STIPPLE_MAX_WIDTH && r.height < STIPPLE_MAX_HEIGHT)
      && r.width != 0 && r.height != 0)
    {
      int i;
      NSScreen *screen = [NSScreen mainScreen];// FIXME/cl for multi-display
      NSSize new;

      for (i = r.width; i < STIPPLE_MAX_WIDTH; i *= 2, new.width = i);
      for (i = r.height; i < STIPPLE_MAX_HEIGHT; i *= 2, new.height = i);

      stippleRep = [[NSImage alloc] initWithSize:new];
      [stippleRep addRepresentation:
                    [[[NSCachedImageRep alloc]
                             initWithSize:[stippleRep size]
                                    depth:[screen depth]
                                 separate:[stippleRep isCachedSeparately]
                                    alpha:YES] autorelease]];
      [stippleRep lockFocus];
      {
        NSPoint p = {0, 0};
        [self compositeToPoint:p operation:NSCompositeCopy];
        if (new.width != r.width)
          {
            NSRect s = { {0, 0}, {0, r.height} };
            NSPoint d = {0, 0};
            for (i = r.width; i != new.width; i *= 2)
              {
                s.size.width = i;
                d.x = i;
                [stippleRep compositeToPoint:d fromRect:s
                                   operation:NSCompositeCopy];
              }
          }
        if (new.height != r.height)
          {
            NSRect s = { {0, 0}, {new.width, 0} };
            NSPoint d = {0, 0};
            for (i = r.height; i != new.height; i *= 2)
              {
                s.size.height = i;
                d.y = i;
                [stippleRep compositeToPoint:d fromRect:s
                                   operation:NSCompositeCopy];
              }
          }
        [stippleRep unlockFocus];
      }
    }
  else
    stippleRep = self;

  return self;
}


- (NSImage *)stippleRep
{
  if (! stippleRep)
    [self prepareForStippling];
  return stippleRep;
}


- initForXPMWithDepth: (int)depth width: (int)width height: (int)height
{
  NSSize s = {width, height};
  id rep;

  [self initWithSize:s];

  rep = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes: NULL
                                  pixelsWide: width pixelsHigh: height
                                  // keep things simple for now
                                  bitsPerSample: 8 samplesPerPixel: 4 // RGB + A
                                  hasAlpha:YES isPlanar:YES
                                  colorSpaceName: NSCalibratedRGBColorSpace
                                  bytesPerRow: width bitsPerPixel: 0];

  [rep getBitmapDataPlanes: xpmData];
  [self addRepresentation:rep];
  return self;
}

// note; this and next assume an image created with initForXPMWithDepth
- (NSColor *) getPixelAtX: (int)x Y: (int)y
{
  if (xpmData == NULL)
    return 0;
  int loc = x + y * [self size].width;
  float r, g, b, a;

  r = xpmData[0][loc] * 256;
  g = xpmData[1][loc] * 256;
  b = xpmData[2][loc] * 256;
  a = xpmData[3][loc] * 256;

  return [NSColor colorWithCalibratedRed: r green: g blue: b alpha: a];
}

- (void) setPixelAtX: (int)x Y: (int)y to: (NSColor *)color
{
  if (xpmData == NULL)
    return;
  int loc = x + y * [self size].width;
  float r, g, b, a;

  [color getRed: &r green: &g blue: &b alpha: &a];

  xpmData[0][loc] = r*255;
  xpmData[1][loc] = g*255;
  xpmData[2][loc] = b*255;
  xpmData[3][loc] = a*255;
}

- (void) setAlphaAtX: (int) x Y: (int) y to: (int) a
{
  if (xpmData == NULL)
    return;

  int loc = x + y * [self size].width;
  xpmData[3][loc] = a;
}

@end
